import logging
from pathlib import Path
from typing import Literal, Union

_PACKAGE_ROOT = Path(__file__).parent
_RESOURCES = _PACKAGE_ROOT / "resources"

VERBOSE = logging.DEBUG - 1  # A level below even debug, logging's lowest

AlternativeSpelling = str
NativeSpelling = str
LanguageMapping = dict[NativeSpelling, AlternativeSpelling]
Span = tuple[int, int]
WordLookup = set[str]
ISO6391Code = Union[
    str,
    # Stupid? Maybe. Useless? Probably.
    Literal[  # https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes
        "aa",
        "ab",
        "ae",
        "af",
        "ak",
        "am",
        "an",
        "ar",
        "as",
        "av",
        "ay",
        "az",
        "ba",
        "be",
        "bg",
        "bi",
        "bm",
        "bn",
        "bo",
        "br",
        "bs",
        "ca",
        "ce",
        "ch",
        "co",
        "cr",
        "cs",
        "cu",
        "cv",
        "cy",
        "da",
        "de",
        "dv",
        "dz",
        "ee",
        "el",
        "en",
        "eo",
        "es",
        "et",
        "eu",
        "fa",
        "ff",
        "fi",
        "fj",
        "fo",
        "fr",
        "fy",
        "ga",
        "gd",
        "gl",
        "gn",
        "gu",
        "gv",
        "ha",
        "he",
        "hi",
        "ho",
        "hr",
        "ht",
        "hu",
        "hy",
        "hz",
        "ia",
        "id",
        "ie",
        "ig",
        "ii",
        "ik",
        "io",
        "is",
        "it",
        "iu",
        "ja",
        "jv",
        "ka",
        "kg",
        "ki",
        "kj",
        "kk",
        "kl",
        "km",
        "kn",
        "ko",
        "kr",
        "ks",
        "ku",
        "kv",
        "kw",
        "ky",
        "la",
        "lb",
        "lg",
        "li",
        "ln",
        "lo",
        "lt",
        "lu",
        "lv",
        "mg",
        "mh",
        "mi",
        "mk",
        "ml",
        "mn",
        "mr",
        "ms",
        "mt",
        "my",
        "na",
        "nb",
        "nd",
        "ne",
        "ng",
        "nl",
        "nn",
        "no",
        "nr",
        "nv",
        "ny",
        "oc",
        "oj",
        "om",
        "or",
        "os",
        "pa",
        "pi",
        "pl",
        "ps",
        "pt",
        "qu",
        "rm",
        "rn",
        "ro",
        "ru",
        "rw",
        "sa",
        "sc",
        "sd",
        "se",
        "sg",
        "si",
        "sk",
        "sl",
        "sm",
        "sn",
        "so",
        "sq",
        "sr",
        "ss",
        "st",
        "su",
        "sv",
        "sw",
        "ta",
        "te",
        "tg",
        "th",
        "ti",
        "tk",
        "tl",
        "tn",
        "to",
        "tr",
        "ts",
        "tt",
        "tw",
        "ty",
        "ug",
        "uk",
        "ur",
        "uz",
        "ve",
        "vi",
        "vo",
        "wa",
        "wo",
        "xh",
        "yi",
        "yo",
        "za",
        "zh",
        "zu",
    ],
]



#!/bin/env python3

"""Tool to replace alternative spellings of native characters
(e.g. German umlauts [ä, ö, ü] etc. [ß]) with the proper native characters.
For example, this problem occurs when no proper keyboard layout was available.

This program is dictionary-based to check if replacements are valid words.

By default, reads from STDIN and writes to STDOUT.
"""

import argparse
import logging
import re
import sys
from typing import TYPE_CHECKING, Any, Iterable, Optional, Union

try:
    import tkinter as tk

    _TKINTER_AVAILABLE = True
except ImportError:
    _TKINTER_AVAILABLE = False

try:
    import pyperclip

    _PYPERCLIP_AVAILABLE = True
except ImportError:
    _PYPERCLIP_AVAILABLE = False

from betterletter import substituters
from betterletter.io import get_dictionary, get_language_mappings
from betterletter.iteration import splitlines

logger = logging.getLogger(__name__)


def parse(
    args: Optional[list[str]], description: str, lang_choices: Iterable[str]
) -> dict[str, Union[bool, str]]:
    """Prepares, runs and returns parsing of CLI arguments for the script."""
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "language",
        help="Text language to work with, in ISO 639-1 format.",
        choices=lang_choices,
    )

    kwargs: dict[str, Union[str, bool]]  # Satisfy mypy
    if _PYPERCLIP_AVAILABLE:
        kwargs = {
            "help": "Read from and write back to clipboard instead of STDIN/STDOUT.",
            "action": "store_true",
        }
    else:
        kwargs = {
            "help": "OPTION IGNORED (not available): package 'pyperclip' couldn't be imported.",
            # False in all cases:
            "action": "store_false",
            "default": False,
        }

    parser.add_argument("-c", "--clipboard", **kwargs)  # type: ignore

    parser.add_argument(
        "-f",
        "--force",
        help="Force substitutions and return the text version with the maximum number"
        " of substitutions, even if they are illegal words (useful for names).",
        action="store_true",
    )
    parser.add_argument(
        "-r",
        "--reverse",
        help="Reverse mode, where all native characters are simply replaced"
        " by their alternative spellings.",
        action="store_true",
    )
    parser.add_argument(
        "-s",
        "--symbols",
        help="Additionally, in a second pass, replace recognized ASCII symbols with their (best-effort) Unicode equivalents (e.g. '->' with '→').",
        action="store_true",
    )

    if _TKINTER_AVAILABLE:
        kwargs = {
            "help": "Stop and open a GUI prompt for confirmation before finishing.",
            "action": "store_true",
        }
    else:
        kwargs = {
            "help": "OPTION IGNORED (not available): package 'tkinter' couldn't be imported.",
            # False in all cases:
            "action": "store_false",
            "default": False,
        }

    parser.add_argument("-g", "--gui", **kwargs)  # type: ignore

    parser.add_argument(
        "-d",
        "--diff",
        help="Print a diff view of the substitutions to stderr.",
        action="store_true",
    )
    parser.add_argument(
        "--debug",
        help="Output detailed logging information.",
        action="store_true",
    )
    return vars(parser.parse_args(args=args))


def main(raw_args: Optional[list[str]] = None) -> None:
    language_mappings = get_language_mappings()

    args = parse(
        args=raw_args, description=__doc__, lang_choices=language_mappings.keys()
    )

    if args["debug"]:
        # Leave at default if no logging/debugging requested.
        logging.basicConfig(level="DEBUG")

    language = str(args["language"])
    language_mapping = language_mappings[language]

    use_clipboard = args["clipboard"]

    if use_clipboard:
        in_text = pyperclip.paste()
        possible_empty_reason = "clipboard empty or binary (document, image, ...)"
    else:
        in_text = sys.stdin.read()
        possible_empty_reason = "STDIN was empty"

    if not in_text.strip():
        logger.warning(f"Empty input received ({possible_empty_reason}).")

    if args["reverse"]:
        out_text = substituters.backward(in_text, language_mapping)
    else:
        try:
            known_words = get_dictionary(language=language)
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Dictionary for {language=} not available (looked for '{e.filename}')"
            ) from e

        out_text = "".join(
            substituters.forward(
                text=in_text,
                language_mapping=language_mapping,
                known_words=known_words,
                force=bool(args["force"]),
            )
        )

    if args["symbols"]:
        out_text = substituters.symbols(out_text)

    if args["diff"] or args["gui"]:
        from difflib import Differ

        raw_diff = Differ().compare(splitlines(in_text), splitlines(out_text))
        # See https://docs.python.org/3/library/difflib.html#difflib.Differ :
        # a 'line common to both sequences' starts with two spaces -> we don't care
        diff = "".join(line for line in raw_diff if not re.match("^  ", line))

    if args["gui"]:
        root = tk.Tk()

        usage = tk.Label(
            text="ENTER to confirm, ESC to abort.",
            font="TkFixedFont",
            fg="white",
            bg="SlateGray",
        )
        usage.pack()

        difflabel = tk.Label(text=diff, font="TkFixedFont", justify="left")
        difflabel.pack()

        root.title("Diff view")

        # https://mypy.readthedocs.io/en/latest/runtime_troubles.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime
        if TYPE_CHECKING:
            Event = tk.Event[Any]
        else:
            Event = tk.Event

        def destroy(event: Event) -> None:
            """Destroys a window and ends mainloop."""
            root.destroy()

        def terminate(event: Event) -> None:
            destroy(event)
            sys.exit(1)

        root.bind("<Return>", destroy)  # Enter key
        root.bind("<Escape>", terminate)

        root.mainloop()

    if args["diff"]:
        sys.stderr.write(diff)

    if use_clipboard:
        pyperclip.copy(out_text)
    else:
        sys.stdout.write(out_text)


if __name__ == "__main__":
    main()



import json
import logging
from functools import cache, partial
from pathlib import Path

from betterletter import _RESOURCES, ISO6391Code, LanguageMapping, WordLookup
from betterletter.iteration import apply_to_all

logger = logging.getLogger(__name__)


# Be explicit about encoding, since Windows might default to something other than UTF8.
# Provide this globally so no spot is forgotten.
_ENCODING = "utf8"
open_with_encoding = partial(open, encoding=_ENCODING)


def read(file: Path) -> list[str]:
    with open_with_encoding(file) as f:
        lines = f.read().splitlines()
    logger.debug(f"Fetched list containing {len(lines)} items from {file}")
    return lines


@cache  # Somewhat heavy io, caching helps a ton for tests etc.
def get_dictionary(language: ISO6391Code) -> WordLookup:
    """Provides the word set from a provided dictionary file.

    Args:
        language: The dictionary of this language will be loaded.

    Returns:
        Set of words from the file.
    """
    path = _RESOURCES / Path("dicts")
    file = Path(language).with_suffix(".txt")

    items = read(path / file)
    logger.debug("Fetched word list.")
    return set(items)


@cache
def get_language_mappings() -> dict[ISO6391Code, LanguageMapping]:
    with open_with_encoding((_RESOURCES / "languages").with_suffix(".json")) as f:
        # Enforce lowercase so we can rely on it later on. Do not use `casefold` on keys,
        # it lowercases too aggressively. E.g., it will turn 'ß' into 'ss', while keys
        # are supposed to be the native letters themselves.
        language_mappings: dict[ISO6391Code, LanguageMapping] = apply_to_all(
            json.load(f), str.lower
        )
    return language_mappings



import logging

logger = logging.getLogger(__name__)


def cf_contains(element: str, string: str) -> bool:
    """Casefold (aka 'strong `lower()`') check if a substring is in a larger string.

    Args:
        element: The shorter string to be tested for containment in `string`.
        string: The larger string.

    Returns:
        Caseless test of whether the larger string contains the shorter string.
    """
    return element.casefold() in string.casefold()
